##
## SciNet's SCMP151, 2023
## Introduction to Quantum Computing
## Erik Spence
##


#########################################################


import pennylane as qml
from pennylane import numpy as np
import pennylane.numpy.random as npr

import sklearn.model_selection as skms
import sklearn.preprocessing as skpp

import pennylane.optimize as po


#########################################################


dev = qml.device("default.qubit", wires = 2)

    
#########################################################

## A function that implements a layer in a quantum circuit,
## implementing some free parameters as arbitrary rotations.
def layer(W):
    
    ## Perform the single-qubit rotations.
    qml.Rot(W[0, 0], W[0, 1], W[0, 2], wires = 0)
    qml.Rot(W[1, 0], W[1, 1], W[1, 2], wires = 1)

    ## Entangle the qubits.
    qml.CNOT(wires = [0, 1])


#########################################################


@qml.qnode(dev, interface = "autograd")
def circuit(weights, x, wires):

    ## Implement the amplitude embedding of the data point.
    qml.AmplitudeEmbedding(x, wires, normalize = True,
                           pad_with = 0)

    ## Add the layers to the circuit.
    for W in weights:
        layer(W)

    ## Return the expectation value of Pauli Z on the first qubit.
    return qml.expval(qml.PauliZ(0))


#########################################################


def cost(weights, features, labels, wires):
    predictions = [circuit(weights, f, wires)
                   for f in features]
    return square_loss(labels, predictions)


#########################################################


def square_loss(labels, predictions):
    loss = 0
    for l, p in zip(labels, predictions):
        loss = loss + (l - p) ** 2

    loss = loss / len(labels)
    return loss


#########################################################


def accuracy(labels, predictions):

    loss = 0
    for l, p in zip(labels, predictions):
        if abs(l - p) < 1e-5:
            loss = loss + 1
    loss = loss / len(labels)

    return loss


#########################################################


def get_data():

    ## Grab the data.
    data = np.loadtxt('occupancy_small.txt', delimiter = ',')

    num_feat = data.shape[1] - 1
    features = data[:, 0:num_feat]

    ## Normalize the feature columns.
    features = skpp.normalize(features, axis = 0)

    ## Grab the targets.
    Y = data[:, -1]

    ## Do a train-test split.
    return skms.train_test_split(features, Y, train_size = 1000,
                                 test_size = 200)
    
    
#########################################################


if __name__ == "__main__":

    print('getting data')
    x_train, x_test, y_train, y_test = get_data()
    
    num_qubits = 2
    num_layers = 4

    wires = range(num_qubits)

    ## Batch size for optimization.
    batch_size = 5
    
    ## Initialize the weights.
    weights = 0.01 * npr.randn(num_layers,
                               num_qubits, 3,
                               requires_grad = True)

    ## Initialize an optimizer.
    opt = po.NesterovMomentumOptimizer(0.01)

    num_train = len(y_train)

    ## Iterate.
    for it in range(101):

        ## Grab a random batch of data.
        batch_index = npr.randint(0, num_train, (batch_size,))
        x_train_batch = x_train[batch_index]
        y_train_batch = y_train[batch_index]

        ## Update the weights by one optimizer step.
        weights, _, _, _ = opt.step(cost, weights,
                                    x_train_batch,
                                    y_train_batch,
                                    wires)

        ## Compute predictions on train and test set
        if (np.mod(it, 10) == 0):

            predictions_train = [np.sign(circuit(weights, f, wires))
                                 for f in x_train]
            predictions_test = [np.sign(circuit(weights, f, wires))
                                for f in x_test]

            # Compute accuracy on train and validation set
            acc_train = accuracy(y_train, predictions_train)
            acc_test = accuracy(y_test, predictions_test)

            print(
                "Iter: {:5d} | Cost: {:0.7f} | Acc train: {:0.7f} | Acc test: {:0.7f} "
                "".format(it, cost(weights, x_train, y_train, wires),
                          acc_train, acc_test)
            )
        

#########################################################
